from pybibs import _internals


def test_split_entries(raw):
    raw_entries = list(_internals.split_entries(raw))
    assert len(raw_entries) == 2


def test_parse_raw_key_values(raw_entry_key_values):
    gen = _internals.parse_raw_key_values(raw_entry_key_values)
    key, val = next(gen)
    assert key == "author"
    assert val == "Israel, Moshe"
    key, val = next(gen)
    assert key == "title"
    assert val == "Some title"


def test_parse_value():
    assert _internals.parse_value(' "Israel, Moshe",\n') == "Israel, Moshe"
    assert _internals.parse_value(" 2008\n") == "2008"



"""
Command line reference manager with a single source of truth: the .bib file.
Inspired by beets.
"""

import os
import pkg_resources
import sys

import click
import click_constraints
import click_plugins
import pybibs
import pyperclip
import requests

from . import cite
from . import internals
from . import query

PATH_OPTION = click.Path(writable=True, readable=True, dir_okay=False)
FILE_OPTIONS = internals.combine_decorators(
    [
        click.option(
            "--file",
            help="File to link to this entry.",
            type=click.Path(exists=True, readable=True, dir_okay=False),
            autocompletion=internals.complete_path,
        ),
        click.option(
            "--destination",
            help="A folder to put the file in.",
            type=click.Path(exists=True, readable=True, dir_okay=True, file_okay=False),
            autocompletion=internals.complete_path,
        ),
        click.option(
            "--no-copy",
            help="Add the specified file in its current location without copying.",
            is_flag=True,
        ),
        click_constraints.constrain("destination", depends=["file"]),
        click_constraints.constrain(
            "no_copy", depends=["file"], conflicts=["destination"],
        ),
    ]
)
SEARCH_TERMS_OPTION = click.argument(
    "search_terms", nargs=-1, autocompletion=internals.complete_key,
)


@click_plugins.with_plugins(pkg_resources.iter_entry_points("bibo.plugins"))
@click.group(help=__doc__)
@click.version_option()
@click.option(
    "--database",
    envvar=internals.BIBO_DATABASE_ENV_VAR,
    help="A .bib file.",
    required=True,
    type=PATH_OPTION,
    autocompletion=internals.complete_path,
)
@click.pass_context
def cli(ctx, database):
    ctx.ensure_object(dict)
    ctx.obj["database"] = os.path.abspath(database)
    ctx.obj["data"] = internals.load_database(database)


@cli.command("list", short_help="List entries.")
@click.option("--raw", is_flag=True, help="Format as raw .bib entries")
@click.option("--bibstyle", default="plain", help="Citation format")
@click.option("--verbose", is_flag=True, help="Print verbose information")
@click.option("--format", help="Format pattern")
@SEARCH_TERMS_OPTION
@click.pass_context
def list_(ctx, search_terms, raw, bibstyle, verbose, **kwargs):
    format_pattern = kwargs.pop("format")
    assert not kwargs

    entries = query.search(ctx.obj["data"], search_terms)
    if raw:
        _list_raw(entries)
    elif format_pattern:
        _list_format_pattern(entries, format_pattern)
    else:
        _list_citations(entries, ctx.obj["database"], bibstyle, verbose)


def _list_raw(entries):
    for entry in entries:
        click.echo(pybibs.write_string([entry]))


def _list_format_pattern(entries, format_pattern):
    for entry in entries:
        click.echo(internals.format_entry(entry, format_pattern))


def _list_citations(entries, database, bibstyle, verbose):
    entries = [e for e in entries if e["type"] != "string"]
    keys = [e["key"] for e in entries]
    exception = None
    try:
        citations = cite.cite(keys, database, bibstyle, verbose)
    except cite.BibtexException as e:
        exception = e

    for entry in entries:
        parts = [
            internals.header(entry),
            cite.fallback(entry) if exception else citations[entry["key"]],
        ]
        click.echo("\n".join(parts))

    if exception is not None:
        parts = [str(exception), "Using a fallback citation method"]
        if exception.use_verbose:
            parts.append("Use --verbose for more information")
        click.secho(". ".join(parts), fg="red")


@cli.command("open", short_help="Open the file / url / doi associated with an entry.")
@SEARCH_TERMS_OPTION
@click.pass_context
def open_(ctx, search_terms):
    entry = query.get(ctx.obj["data"], search_terms)

    for field_name in ["file", "url", "doi"]:
        value = entry.get("fields", {}).get(field_name)
        if value:
            if field_name == "doi":
                value = "https://doi.org/" + value
            click.launch(value)
            break
    else:
        raise click.ClickException("No file / url / doi is associated with this entry")


@cli.command(short_help="Add a new entry.")
@FILE_OPTIONS
@click.option("--doi", help="Add entry by DOI.")
@click.pass_context
def add(ctx, destination, doi, no_copy, **kwargs):
    file_ = kwargs.pop("file")

    data = ctx.obj["data"]
    if doi is not None:
        url = "http://dx.doi.org/{}".format(doi)
        headers = {"Accept": "application/x-bibtex"}
        resp = requests.get(url, headers=headers)
        assert resp.status_code == 200
        raw_bib = resp.text
    else:
        raw_bib = pyperclip.paste()
    bib = internals.editor(text=raw_bib)
    entry = pybibs.read_entry_string(bib)

    internals.unique_key_validation(entry["key"], data)

    data.append(entry)

    if file_:
        internals.set_file(data, entry, file_, destination, no_copy)

    pybibs.write_file(data, ctx.obj["database"])


@cli.command(short_help="Remove an entry or a field.")
@click.argument("key", autocompletion=internals.complete_key)
@click.argument("field", nargs=-1)
@click.pass_context
def remove(ctx, key, field):
    data = ctx.obj["data"]
    entry = query.get_by_key(data, key)

    if not field:
        data.remove(entry)
    elif "fields" in entry:
        for f in field:
            if f in entry["fields"]:
                del entry["fields"][f]
            else:
                click.echo('"{}" has no field "{}"'.format(key, f))
    else:
        click.echo('"{}" has no fields'.format(key))

    pybibs.write_file(data, ctx.obj["database"])


@cli.command(short_help="Edit an entry.")
@click.argument("key", autocompletion=internals.complete_key)
@click.argument("field_value", nargs=-1)
@FILE_OPTIONS
@click.pass_context
def edit(ctx, key, field_value, destination, no_copy, **kwargs):
    """
    FIELD_VALUE

    To set fields: author=Einstein tags=interesting.
    Leave the value empty to open in editor.
    Set the key / type in the same way.
    """
    file_ = kwargs.pop("file")

    data = ctx.obj["data"]
    entry = query.get_by_key(data, key)

    if file_:
        internals.set_file(data, entry, file_, destination, no_copy)
    for fv in field_value:
        if "=" in fv:
            field, value = fv.split("=")
        else:
            field = fv
            current_value = entry["fields"].get(field, "")
            value = internals.editor(text=current_value).strip()
        if field == "key":
            internals.unique_key_validation(value, data)
            entry["key"] = value
        elif field == "type":
            entry["type"] = value
        else:
            entry["fields"][field] = value

    pybibs.write_file(data, ctx.obj["database"])


if __name__ == "__main__":
    cli()



# coding=utf-8

import collections
import glob
import itertools
import os
import re
import shutil

import click

import pybibs

BIBO_DATABASE_ENV_VAR = "BIBO_DATABASE"


def header(entry):
    parts = [click.style(entry["key"], fg="green")]
    fields = entry["fields"]
    if fields.get("tags"):
        parts.append(click.style(fields["tags"], fg="cyan"))
    if fields.get("file"):
        parts.append("📁")
    if fields.get("url") or fields.get("doi"):
        parts.append("🔗")
    return " ".join(parts)


def format_entry(entry, format_pattern):
    output = []
    replacement_start_index = -1
    for i, char in enumerate(format_pattern):
        if char == "$":
            replacement_start_index = i + 1
        elif (not char.isalpha()) and (replacement_start_index >= 0):
            field = format_pattern[replacement_start_index:i]
            output.append(_lookup(entry, field))
            output.append(char)
            replacement_start_index = -1
        elif replacement_start_index == -1:
            output.append(char)
    if replacement_start_index >= 0:
        field = format_pattern[replacement_start_index:]
        output.append(_lookup(entry, field))
    return "".join(output)


def _lookup(entry, field):
    if field in entry:
        return entry[field]
    if "fields" in entry and field in entry["fields"]:
        return entry["fields"][field]
    return "$" + field


def destination_heuristic(data):
    """
    A heuristic to get the folder with all other files from bib, using majority
    vote.
    """
    counter = collections.Counter()
    for entry in bib_entries(data):
        file_field = entry["fields"].get("file")
        if not file_field:
            continue
        path = os.path.dirname(file_field)
        counter[path] += 1

    if not counter:  # No paths found
        raise click.ClickException(
            "Path finding heuristics failed: no paths in the database"
        )

    # Find the paths that appears most often
    sorted_paths = sorted(counter, reverse=True)
    groupby = itertools.groupby(sorted_paths, key=len)
    _, group = next(groupby)

    # We know that there's at least one candidate. Make sure it's
    # the only one
    candidate = next(group)
    try:
        next(group)
    except StopIteration:
        return candidate
    else:
        raise click.ClickException(
            "Path finding heuristics failed: "
            "there are multiple equally valid paths in the database"
        )


def string_to_basename(s):
    """
    Converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """
    s = s.strip().lower()
    s = re.sub(r"[^\w\s-]", "", s)
    return re.sub(r"[\s-]+", "-", s)


def set_file(data, entry, file_, destination=None, no_copy=False):
    if no_copy:
        entry["fields"]["file"] = os.path.abspath(file_)
        return

    if not destination:
        destination = destination_heuristic(data)
    destination = os.path.abspath(destination)
    _, file_extension = os.path.splitext(file_)
    basename = string_to_basename(entry["key"])
    path = os.path.join(destination, basename + file_extension)
    entry["fields"]["file"] = path
    click.echo("Copying {} to {}".format(file_, path))
    shutil.copy(file_, path)


def editor(*args, **kwargs):
    """
    Wrapper for `click.edit` that raises an error when None is returned.
    """
    result = click.edit(*args, **kwargs)
    if result is None:
        msg = "Editor exited without saving, command aborted"
        raise click.ClickException(msg)
    return result


def get_database(args):
    for arg_a, arg_b in zip(args, args[1:]):
        if arg_a == "--database":
            return arg_b


def complete_key(ctx, args, incomplete):
    """
    Autocompletion for keys.
    """
    database = get_database(args) or os.environ.get(BIBO_DATABASE_ENV_VAR)
    if database:
        data = load_database(database)
    else:
        data = []
    data = bib_entries(data)
    return [x["key"] for x in data if incomplete.lower() in x["key"].lower()]


def load_database(database):
    """
    Load the database from .bib. Create (in memory) if doesn't exist.
    """
    try:
        return pybibs.read_file(database)
    except IOError:
        return []


def combine_decorators(decorators):
    # Copied from https://stackoverflow.com/a/4122845/1224456
    def decorator(f):
        for d in reversed(decorators):
            f = d(f)
        return f

    return decorator


def bib_entries(entries):
    """
    Yield only the actual bibliographic items from a list of entries.
    Drop @string / @comment / @preamble entries.
    """
    for e in entries:
        if e["type"] not in ["string", "comment", "preamble"]:
            yield e


def unique_key_validation(new_key, data):
    if new_key in (e["key"] for e in bib_entries(data)):
        raise click.ClickException("Duplicate key, command aborted")


def complete_path(ctx, args, incomplete):
    """
    Autocompletion for files, matches the glob of incomplete argument, and
    provide basename prompts.
    """
    wildc_path = os.path.expanduser(os.path.expandvars(incomplete)) + "*"
    options = []
    for path in glob.glob(wildc_path):
        if os.access(path, os.R_OK):
            options.append((path, os.path.basename(path)))
    return options



import os

import pybibs


def test_read_string(raw):
    bib = pybibs.read_string(raw)
    assert len(bib) == 2
    entry = bib[0]
    assert entry["key"] == "israel"
    assert entry["type"] == "article"
    assert entry["fields"]["author"] == "Israel, Moshe"
    assert entry["fields"]["year"] == "2008"


def test_read_file(database):
    bib = pybibs.read_file(database)
    for entry in bib:
        if entry.get("key") == "bailenson2005digital":
            break
    else:
        raise AssertionError("Couldn't find bailenson2005digital in parsed data")


def test_entry_contains_key_and_type(raw):
    bib = pybibs.read_string(raw)
    assert bib[0]["key"] == "israel"
    assert bib[0].get("type") == "article"


def test_read_multiline(multiline_entry):
    bib = pybibs.read_string(multiline_entry)
    assert bib[0]["fields"]["author"] == "Israel, Moshe and Yosef, Shlomo"


def test_read_bens_multiline(bens_multiline_entry):
    bib = pybibs.read_string(bens_multiline_entry)
    fields = bib[0]["fields"]
    expected_title = "{Radio Time-Domain Signatures of Magnetar Birth}"
    expected_keywords = "Astrophysics - High Energy Astrophysical Phenomena"
    assert fields["title"] == expected_title
    assert fields["keywords"] == expected_keywords
    assert fields["year"] == "2019"


def test_string_type_entry():
    raw = '@string{foo = "Mrs. Foo"}'
    bib = pybibs.read_string(raw)
    assert bib[0]["type"] == "string"
    assert bib[0]["key"] == "foo"
    assert bib[0]["val"] == "Mrs. Foo"


def test_comment_type_entry():
    raw = "@comment{This is a comment}"
    bib = pybibs.read_string(raw)
    assert bib[0] == {"type": "comment", "body": "This is a comment"}


def test_preamble_type_entry():
    raw = r'@preamble{"Some \latex code"}'
    bib = pybibs.read_string(raw)
    assert bib[0] == {"type": "preamble", "body": r'"Some \latex code"'}
